from typing import Union

from ..sqlalchemy_ import sa

from autofit.mapper.prior import abstract
from autofit.mapper.prior_model import prior_model
from autofit.mapper.prior_model import collection

from .model import Object


class CollectionPriorModel(Object):
    """
    A collection
    """

    __tablename__ = "collection_prior_model"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True,
    )

    __mapper_args__ = {
        'polymorphic_identity': 'collection_prior_model'
    }

    @classmethod
    def _from_object(
            cls,
            source: Union[
                collection.CollectionPriorModel,
                list,
                dict
            ]
    ):
        instance = cls()
        if not isinstance(
                source,
                collection.CollectionPriorModel
        ):
            source = collection.CollectionPriorModel(
                source
            )
        instance._add_children(
            source.items()
        )
        instance.cls = collection.CollectionPriorModel
        return instance


class PriorModel(Object):
    """
    A prior model
    """

    __tablename__ = "prior_model"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True,
    )

    __mapper_args__ = {
        'polymorphic_identity': 'prior_model'
    }

    @classmethod
    def _from_object(
            cls,
            model: prior_model.PriorModel,
    ):
        instance = cls()
        instance.cls = model.cls
        instance._add_children(model.items())
        return instance

    def _make_instance(self):
        instance = object.__new__(prior_model.PriorModel)
        instance.cls = self.cls
        instance._assertions = []
        return instance


class Prior(Object):
    """
    A prior
    """

    __tablename__ = "prior"

    id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        ),
        primary_key=True,
    )

    __mapper_args__ = {
        'polymorphic_identity': 'prior'
    }

    @classmethod
    def _from_object(
            cls,
            model: abstract.Prior
    ):
        instance = cls()
        instance.cls = type(model)
        instance._add_children([
            (key, value)
            for key, value
            in model.__dict__.items()
            if key in model.__database_args__
        ])
        return instance

    def __call__(self):
        """
        Create the real instance for this object, with child
        attributes attached.

        If the instance implements __setstate__ then this is
        called with a dictionary of instantiated children.
        """
        arguments = {
            child.name: child()
            for child in self.children
        }
        return self.cls(
            **arguments
        )



from autoconf.exc import PriorException


class MessageException(PriorException):
    """
    Raised when some assertion about the parameterization of a message is not met
    """


class PathsException(Exception):
    pass


class FitException(Exception):
    """
    An exception to be thrown if the non linear search must resample; equivalent to returning an infinitely bad fit
    """

    pass


class PriorLimitException(FitException, PriorException):
    pass


class PipelineException(Exception):
    pass


class DeferredInstanceException(Exception):
    """
    Exception raised when an attempt is made to access an attribute or function of a
    deferred instance prior to instantiation
    """

    pass


class AggregatorException(Exception):
    pass


class GridSearchException(Exception):
    pass


class HistoryException(Exception):
    """
    Thrown when insufficient factor history is present for a given operation
    """

class SamplesException(Exception):
    pass


class SamplesWarning(Warning):
    pass


from autofit.non_linear.mock.mock_analysis import MockAnalysis
from autofit.non_linear.mock.mock_result import MockResult
from autofit.non_linear.mock.mock_result import MockResultGrid
from autofit.non_linear.mock.mock_search import MockSearch
from autofit.non_linear.mock.mock_search import MockOptimizer
from autofit.non_linear.mock.mock_samples import MockSamples
from autofit.non_linear.mock.mock_samples import MockNestSamples

from autofit.mapper.mock.mock_model import MockChildTuple
from autofit.mapper.mock.mock_model import MockClassInf
from autofit.mapper.mock.mock_model import MockClassRelativeWidth
from autofit.mapper.mock.mock_model import MockChildTuplex3
from autofit.mapper.mock.mock_model import MockChildTuplex2
from autofit.mapper.mock.mock_model import MockClassx3TupleFloat
from autofit.mapper.mock.mock_model import MockClassx2
from autofit.mapper.mock.mock_model import MockClassx2Instance
from autofit.mapper.mock.mock_model import MockClassx2FormatExp
from autofit.mapper.mock.mock_model import MockClassx2NoSuperScript
from autofit.mapper.mock.mock_model import MockClassx2Tuple
from autofit.mapper.mock.mock_model import MockClassx3
from autofit.mapper.mock.mock_model import MockClassx3TupleFloat
from autofit.mapper.mock.mock_model import MockClassx4
from autofit.mapper.mock.mock_model import MockComplexClass
from autofit.mapper.mock.mock_model import MockComponents
from autofit.mapper.mock.mock_model import MockDeferredClass
from autofit.mapper.mock.mock_model import MockListClass
from autofit.mapper.mock.mock_model import MockOverload
from autofit.mapper.mock.mock_model import MockParent
from autofit.mapper.mock.mock_model import MockWithFloat
from autofit.mapper.mock.mock_model import MockWithTuple




from copy import copy
from typing import Union, Type, Optional, Tuple, Iterator

import numpy as np

from autofit.messages.transform import AbstractDensityTransform, LinearShiftTransform
from .abstract import AbstractMessage


class TransformedWrapperInstance:
    """
    An instance of a transformed message. e.g. a UniformNormal message.

    This allows arbitrary transforms to be created on the fly while supporting
    the same interface. The true underlying message is still the same, but values
    computed from it are transformed to give the effect of a different distribution.
    """

    def value_for(self, unit: float) -> float:
        return self.instance().value_for(unit)

    def __init__(
            self,
            transformed_wrapper: "TransformedWrapper",
            *args,
            **kwargs
    ):
        """
        Parameters
        ----------
        transformed_wrapper
            Acts like a message class but provides a transformation on
            some underlying class.
        args
            Arguments required to instantiate the underlying message class
        kwargs
            Keyword arguments required to instantiate the underlying message
            class
        """
        self.id = kwargs.get("id_")
        self.transformed_wrapper = transformed_wrapper
        self.args = args
        self.kwargs = kwargs

        self._instance = None

    def as_message(self):
        return self._new_for_base_message(self)

    def project(
            self,
            samples: np.ndarray,
            log_weight_list: Optional[np.ndarray] = None,
            **kwargs,
    ):
        return self._new_for_base_message(
            self.transformed_wrapper.project(
                samples=samples,
                log_weight_list=log_weight_list,
                **kwargs,
            )
        )

    def _new_for_base_message(
            self,
            message
    ):
        """
        Create a new instance of this wrapper but change the parameters used
        to instantiate the underlying message. This is useful for retaining
        the same transform stack after recreating the underlying message during
        projection.
        """
        return type(self)(
            self.transformed_wrapper,
            *message.parameters,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
            id_=self.instance().id,
        )

    def __mul__(self, other):
        """
        Multiply this message by some other message. Effectively multiplies the
        underlying message whilst retaining the same transform stack.
        """
        if isinstance(
                other,
                TransformedWrapperInstance
        ):
            other = other.instance()
        return self._new_for_base_message(
            self.instance() * other
        )

    def __pow__(self, other):
        """
        Compute this message to the power of some value/
        """
        if isinstance(
                other,
                TransformedWrapperInstance
        ):
            other = other.instance()
        return self._new_for_base_message(
            self.instance() * other
        )

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        """
        Divide this message by some other message. Effectively divides the
        underlying message whilst retaining the same transform stack.
        """
        return self._new_for_base_message(
            self.instance() / other.instance()
        )

    def __sub__(self, other):
        instance = self.instance()
        if isinstance(
                other,
                TransformedWrapperInstance
        ):
            other = other.instance()
        return instance - other

    def __eq__(self, other):
        if not isinstance(
                other,
                TransformedWrapperInstance
        ):
            return False
        return other.instance() == self.instance()

    def __getattr__(self, item):
        """
        By default attributes are taken from the underlying message instance
        """
        return getattr(
            self.instance(),
            item
        )

    def update_invalid(self, other: "AbstractMessage") -> "TransformedWrapperInstance":
        valid = self.check_valid()
        if self.ndim:
            valid_parameters: Iterator[np.ndarray] = (
                np.where(valid, p, p_safe) for p, p_safe in zip(self, other)
            )
        else:
            # TODO: Fairly certain this would not work
            valid_parameters = iter(self if valid else other)
        return TransformedWrapperInstance(
            self.transformed_wrapper,
            *valid_parameters,
            **self.kwargs,
        )

    def __iter__(self):
        return iter(self.instance())

    def copy(self):
        copied = copy(self)
        self._instance = None
        return copied

    @property
    def cls(self):
        return self.transformed_wrapper.cls

    def __hash__(self):
        return hash(self.instance())

    def __setstate__(self, state):
        self._instance = None
        self.__dict__.update(state)

    def __getstate__(self):
        """
        Representation of state of object for pickling excluding
        underlying instance as this can be reconstructed.
        """
        return {
            key: value
            for key, value
            in self.__dict__.items()
            if "_instance" != key
        }

    def instance(self):
        """
        An instance of the transformed message wrapped by this class.
        """
        if self._instance is None:
            cls = self.transformed_wrapper.transformed_class()
            self._instance = cls(
                *self.args,
                **self.kwargs,
            )
            self._instance.id = self.id
        return self._instance

    @property
    def factor(self):
        return self.instance().factor

    def from_mode(
            self,
            mode: np.ndarray,
            covariance: np.ndarray,
            **kwargs
    ):
        return self._new_for_base_message(
            self.transformed_wrapper.from_mode(
                mode,
                covariance,
                **kwargs
            )
        )


class TransformedWrapper:
    """
    A transformed message. This allows transformed messages to be created
    on the fly whilst retaining the same API.
    """

    InstanceWrapper = TransformedWrapperInstance

    def __init__(
            self,
            cls: Union[Type[AbstractMessage], "TransformedWrapper"],
            transform: AbstractDensityTransform,
            clsname: Optional[str] = None,
            support: Optional[Tuple[Tuple[float, float], ...]] = None,
    ):
        """
        Parameters
        ----------
        cls
            The class or TransformWrapper being transformed
        transform
            The transform applied
        clsname
            A custom name for the newly created class
        support
            The supported region
        """
        self.cls = cls
        self.transform = transform
        self.clsname = clsname
        self.support = support

        self.__transformed_class = None

    def __getattr__(self, item):
        """
        By default attributes values are taken from the underlying message class
        """
        return getattr(
            self.transformed_class(),
            item
        )

    def __call__(self, *args, **kwargs):
        """
        'Instantiate' the message. Creates an instance wrapper and
        passes arguments and keyword arguments to it for lazy instantiation
        of the underlying message.
        """
        return self.InstanceWrapper(
            self,
            *args,
            **kwargs
        )

    def transformed(
            self,
            transform: Union[
                AbstractDensityTransform,
                Type[AbstractDensityTransform]
            ],
            clsname: Optional[str] = None,
            support: Optional[Tuple[Tuple[float, float], ...]] = None,
            wrapper_cls=None
    ) -> "TransformedWrapper":
        """
        Apply a further transformation to an already transformed message
        """
        wrapper_cls = wrapper_cls or TransformedWrapper
        return wrapper_cls(
            cls=self,
            transform=transform,
            clsname=clsname,
            support=support,
        )

    def shifted(
            self,
            shift: float = 0,
            scale: float = 1,
            wrapper_cls=None,
    ):
        return self.transformed(
            LinearShiftTransform(shift=shift, scale=scale),
            clsname=f"Shifted{self.transformed_class().__name__}",
            wrapper_cls=wrapper_cls
        )

    def __getstate__(self):
        """
        The definition of the transformed class is not included in the state
        to circumvent an issue with pickling arbitrarily parameterised on-the-fly
        class definitions.

        The definition of the class is lazily created from the state of an instance
        as required.
        """
        return {
            key: value
            for key, value
            in self.__dict__.items()
            if "__transformed_class" not in key
        }

    def transformed_class(self):
        """
        The transformed class is lazily created
        """
        if self.__transformed_class is None:
            self.__transformed_class = self._transformed_class()
        return self.__transformed_class

    def _transformed_class(self):
        """
        The transformed class is lazily defined. This means it does not have to be
        pickled; it can entirely be derived from the state of the TransformedWrapper
        """
        from .transformed import TransformedMessage

        if isinstance(self.cls, TransformedWrapper):
            depth = self.cls._depth + 1
            clsname = self.clsname or f"Transformed{depth}{self.cls._Message.__name__}"

            # Don't doubly inherit if transforming already transformed message
            cls = self.cls.transformed_class()

            class Transformed(cls):
                __qualname__ = clsname
                _depth = depth
                _Message = cls
        else:
            cls = self.cls
            clsname = self.clsname or f"Transformed{self.cls.__name__}"

            class Transformed(TransformedMessage):  # type: ignore
                __qualname__ = clsname
                parameter_names = self.cls.parameter_names
                _depth = 1

        projectionClass = (
            None if cls._projection_class is None
            else cls._projection_class.transformed(self.transform)
        )

        support = self.support or tuple(zip(*map(
            self.transform.inv_transform, map(np.array, zip(*cls._support))
        ))) if cls._support else cls._support

        Transformed._transform = self.transform
        Transformed._support = support
        Transformed.__projection_class = projectionClass
        Transformed.__name__ = clsname
        Transformed._Message = cls

        return Transformed

    def __setstate__(self, state):
        self.__transformed_class = None
        self.__dict__.update(state)

    def from_mode(
            self,
            mode: np.ndarray,
            covariance: np.ndarray,
            **kwargs
    ):
        mode, jac = self._transform.transform_jac(mode)
        if covariance.shape != ():
            covariance = jac.quad(covariance)
        return self.cls.from_mode(mode, covariance, **kwargs)



from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.collection import CollectionPriorModel
from .analysis import Analysis
from .indexed import IndexCollectionAnalysis


class ModelAnalysis(Analysis):
    def __init__(
            self,
            analysis: Analysis,
            model: AbstractPriorModel
    ):
        """
        Comprises a model and an analysis that can be applied to instances of that model.

        Parameters
        ----------
        analysis
        model
        """
        self.analysis = analysis
        self.model = model

    def __getattr__(self, item):
        return getattr(self.analysis, item)

    def log_likelihood_function(self, instance):
        return self.analysis.log_likelihood_function(instance)


class CombinedModelAnalysis(IndexCollectionAnalysis):
    def modify_model(
            self,
            model: AbstractPriorModel
    ) -> CollectionPriorModel:
        """
        Creates a collection with one model for each analysis. For each ModelAnalysis
        the model is used; for other analyses the default model is used.

        Parameters
        ----------
        model
            A default model

        Returns
        -------
        A collection of models, one for each analysis.
        """
        return CollectionPriorModel([
            analysis.analysis.model if isinstance(
                analysis.analysis,
                ModelAnalysis
            ) else model
            for analysis in self.analyses
        ])



import inspect
import logging
import sys

from autoconf import conf
from autoconf.exc import ConfigException

logger = logging.getLogger(
    __name__
)


class WidthModifier:
    def __init__(self, value):
        self.value = float(value)

    @classmethod
    def name_of_class(cls) -> str:
        """
        A string name for the class, with the prior suffix removed.
        """
        return cls.__name__.replace("WidthModifier", "")

    @classmethod
    def from_dict(cls, width_modifier_dict):
        return width_modifier_type_dict[width_modifier_dict["type"]](
            value=width_modifier_dict["value"]
        )

    @property
    def dict(self):
        return {"type": self.name_of_class(), "value": self.value}

    @staticmethod
    def for_class_and_attribute_name(
            cls: type,
            attribute_name: str
    ) -> "WidthModifier":
        """
        Search prior configuration for a WidthModifier.

        If no configuration is found a RelativeWidthModifier
        with a value of 0.5 is returned.

        Parameters
        ----------
        cls
            The class to which an attribute belongs
        attribute_name
            The name of an attribute

        Returns
        -------
        An object that ensures the prior derived from the
        posterior of a search is sufficiently wide.
        """
        try:
            prior_dict = conf.instance.prior_config.for_class_and_suffix_path(
                cls, [attribute_name, "width_modifier"]
            )
            return WidthModifier.from_dict(prior_dict)
        except ConfigException:
            logger.warning(
                f"No width modifier specified for class:\n\n"
                f"{cls.__name__}\n\n"
                f"and attribute:\n\n"
                f"{attribute_name}\n\n"
                f"Using default relative modifier of 0.5. See "
                f"https://pyautofit.readthedocs.io/en/latest/general/adding_a_model_component.html"
                f" for more information."
            )
            return RelativeWidthModifier(0.5)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.value == other.value


class RelativeWidthModifier(WidthModifier):
    def __call__(self, mean):
        return self.value * mean


class AbsoluteWidthModifier(WidthModifier):
    def __call__(self, _):
        return self.value


width_modifier_type_dict = {
    obj.name_of_class(): obj
    for _, obj in inspect.getmembers(sys.modules[__name__])
    if inspect.isclass(obj) and issubclass(obj, WidthModifier)
}



import logging
from abc import ABC, abstractmethod
from typing import Tuple, Optional

from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.mean_field import MeanField, FactorApproximation, Status
from autofit.graphical.utils import LogWarnings

logger = logging.getLogger(__name__)


class AbstractFactorOptimiser(ABC):
    """
    An optimiser used to optimise individual factors during EPOptimisation.
    """
    logger = logger.debug

    def __init__(self, initial_values=None, deltas=None, inplace=False, delta=1):
        self.initial_values = initial_values or {}
        self.inplace = inplace
        self.delta = delta
        self.deltas = deltas or {}

    def update_model_approx(
            self,
            new_model_dist: MeanField,
            factor_approx: FactorApproximation,
            model_approx: EPMeanField,
            status: Optional[Status] = Status(),
            delta: Optional[float] = None,
    ) -> Tuple[EPMeanField, Status]:
        delta = delta or self.deltas.get(factor_approx.factor) or self.delta
        new_approx, status = model_approx.project_mean_field(
            new_model_dist,
            factor_approx,
            delta=delta,
            status=status,
        )
        return new_approx, status

    @abstractmethod
    def optimise(
            self, factor: Factor, model_approx: EPMeanField, status: Status = Status()
    ) -> Tuple[EPMeanField, Status]:
        pass

    def exact_fit(
            self, factor: Factor, model_approx: EPMeanField, status: Status = Status()
    ) -> Tuple[EPMeanField, Status]:

        with LogWarnings(logger=self.logger, action='always') as caught_warnings:
            if factor._calc_exact_update:
                factor_approx = model_approx.factor_approximation(factor)
                new_approx = model_approx if self.inplace else model_approx.copy()
                new_approx.update_factor_mean_field(
                    factor, factor.calc_exact_update(factor_approx.cavity_dist)
                )
            elif factor._calc_exact_projection:
                factor_approx = model_approx.factor_approximation(factor)
                new_model_dist = factor.calc_exact_projection(factor_approx.cavity_dist)
                new_approx, status = self.update_model_approx(
                    new_model_dist, factor_approx, model_approx, status
                )

            else:
                raise NotImplementedError(
                    "Factor does not have exact updates methods"
                )

        status_kws = status._asdict()
        status_kws['messages'] = status.messages + tuple(caught_warnings.messages)
        status = Status(**status_kws)

        return new_approx, status


class ExactFactorFit(AbstractFactorOptimiser):
    optimise = AbstractFactorOptimiser.exact_fit



import os
from os import path
import matplotlib.pyplot as plt
import numpy as np
from typing import Optional

def plot_profile_1d(
    xvalues : np.ndarray,
    profile_1d: np.ndarray,
    title:Optional[str]=None,
    ylabel:Optional[str]=None,
    errors:Optional[np.ndarray]=None,
    color:Optional[str]="k",
    output_path:Optional[str]=None,
    output_filename:Optional[str]=None,
):
    """
    Plot a 1D image of data on a plot of x versus y, where the x-axis is the x coordinate of the 1D profile
    and the y-axis is the value of the 1D profile at that coordinate.

    The function include options to output the image to the hard-disk as a .png.

    Parameters
    ----------
    xvalues
        The x-coordinates the profile is defined on.
    profile_1d
        The normalization values of the profile which are plotted.
    ylabel
        The y-label of the plot.
    errors
        The errors on each data point, which are related to its noise-map.
    output_path
        The path the image is to be output to hard-disk as a .png.
    output_filename
        The filename of the file if it is output as a .png.
    output_format
        Determines where the plot is displayed on your screen ("show") or output to the hard-disk as a png ("png").
    """
    plt.errorbar(
        x=xvalues, y=profile_1d, yerr=errors, color=color, ecolor="k", elinewidth=1, capsize=2
    )
    plt.title(title)
    plt.xlabel("x value of profile")
    plt.ylabel(ylabel)
    if output_filename is None:
        plt.show()
    else:
        if not path.exists(output_path):
            os.makedirs(output_path)
        plt.savefig(path.join(output_path, f"{output_filename}.png"))
    plt.clf()


import math
import numpy as np
from typing import List, Optional
import warnings

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.samples.pdf import PDFSamples
from autofit.non_linear.samples.samples import Samples
from autofit.non_linear.samples.samples import Sample
from autofit.non_linear.samples.sample import load_from_table

from autofit import exc

class MCMCSamples(PDFSamples):

    def __init__(
            self,
            model: AbstractPriorModel,
            sample_list: List[Sample],
            auto_correlation_settings: AutoCorrelationsSettings,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
            results_internal: Optional = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the results_internal of
        a `NonLinearSearch` (e.g. as files on your hard-disk) and Python.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        This class stores the samples of a MCMC model-fit (e.g. `emcee`, `zeus`). To use a library's in-built
        visualization tools results_internal are optionally stored in their native internal format using
        the `results_internal` attribute.

        Attributes
        ----------
        results_internal
            The MCMC results in their native internal format from which the samples are computed.
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        auto_correlations_settings
            Customizes and performs auto correlation calculations performed during and after the search.
        unconverged_sample_size
            If the samples are for a search that is yet to convergence, a reduced set of samples are used to provide
            a rough estimate of the parameters. The number of samples is set by this parameter.
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        results_internal
            The MCMC library's results in their native internal format for interfacing its visualization library.
        """

        self.auto_correlation_settings = auto_correlation_settings

        super().__init__(
            model=model,
            sample_list=sample_list,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
            results_internal=results_internal
        )

    def __add__(
            self,
            other: "MCMCSamples"
    ) -> "MCMCSamples":
        """
        Samples can be added together, which combines their `sample_list` meaning that inferred parameters are
        computed via their joint PDF.

        For Zeus samples there are no tools for combining results in their native format, therefore these
        `results_internal` are set to None and support for visualization is disabled.

        Parameters
        ----------
        other
            Another Samples class

        Returns
        -------
        A class that combined the samples of the two Samples objects.
        """

        self._check_addition(other=other)

        warnings.warn(
            f"Addition of {self.__class__.__name__} cannot retain results in native format. "
            "Visualization of summed samples diabled.",
            exc.SamplesWarning
        )

        return self.__class__(
            model=self.model,
            sample_list=self.sample_list + other.sample_list,
            auto_correlation_settings=self.auto_correlation_settings,
            unconverged_sample_size=self.unconverged_sample_size,
            time=self.time,
            results_internal=None
        )

    @property
    def total_walkers(self):
        raise NotImplementedError

    @property
    def total_steps(self):
        raise NotImplementedError

    @property
    def auto_correlations(self):
        raise NotImplementedError

    @classmethod
    def from_table(self, filename: str, model: AbstractPriorModel, number_live_points: int = None):
        """
        Write a table of parameters, posteriors, priors and likelihoods

        Parameters
        ----------
        filename
            Where the table is to be written
        """

        sample_list = load_from_table(filename=filename)

        return Samples(
            model=model,
            sample_list=sample_list
        )

    @property
    def info_json(self):
        return {
            "times": None,
            "check_size": self.auto_correlations.check_size,
            "required_length": self.auto_correlations.required_length,
            "change_threshold": self.auto_correlations.change_threshold,
            "total_walkers": self.total_walkers,
            "total_steps": self.total_steps,
            "time": self.time,
        }

    @property
    def pdf_converged(self):
        """
        To analyse and visualize samples using corner.py, the analysis must be sufficiently converged to produce
        smooth enough PDF for analysis. This property checks whether the non-linear search's samples are sufficiently
        converged for corner.py use.

        Emcee samples can be analysed by corner.py irrespective of how long the sampler has run, albeit low run times
        will likely produce inaccurate results.
        """
        try:
            samples_after_burn_in = self.samples_after_burn_in
            if len(samples_after_burn_in) == 0:
                return False
            return True
        except ValueError:
            return False

    @property
    def samples_after_burn_in(self) -> [List]:
        """
        The emcee samples with the initial burn-in samples removed.

        The burn-in period is estimated using the auto-correlation times of the parameters.
        """
        raise NotImplementedError()

    @property
    def converged(self) -> bool:
        """
        Whether the emcee samples have converged on a solution or if they are still in a burn-in period, based on the
        auto correlation times of parameters.
        """
        return self.auto_correlations.check_if_converged(
            total_samples=self.total_samples
        )

    @property
    def median_pdf_vector(self) -> [float]:
        """
        The median of the probability density function (PDF) of every parameter marginalized in 1D, returned
        as a list of values.

        This is computed by binning all sampls after burn-in into a histogram and take its median (e.g. 50%) value.
        """

        if self.pdf_converged:
            return [
                float(np.percentile(self.samples_after_burn_in[:, i], [50]))
                for i in range(self.model.prior_count)
            ]

        return self.max_log_likelihood_vector

    def vector_at_sigma(self, sigma: float) -> [float]:
        """
        The value of every parameter marginalized in 1D at an input sigma value of its probability density function
        (PDF), returned as two lists of values corresponding to the lower and upper values parameter values.

        For example, if sigma is 1.0, the marginalized values of every parameter at 31.7% and 68.2% percentiles of each
        PDF is returned.

        This does not account for covariance between parameters. For example, if two parameters (x, y) are degenerate
        whereby x decreases as y gets larger to give the same PDF, this function will still return both at their
        upper values. Thus, caution is advised when using the function to reperform a model-fits.

        For Emcee, if the samples have converged this is estimated by binning the samples after burn-in into a
        histogram and taking the parameter values at the input PDF %.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        limit = math.erf(0.5 * sigma * math.sqrt(2))

        if self.pdf_converged:
            samples = self.samples_after_burn_in

            return [
                tuple(
                    np.percentile(samples[:, i], [100.0 * (1.0 - limit), 100.0 * limit])
                )
                for i in range(self.model.prior_count)
            ]

        parameters_min = list(
            np.min(self.parameter_lists[-self.unconverged_sample_size:], axis=0)
        )
        parameters_max = list(
            np.max(self.parameter_lists[-self.unconverged_sample_size:], axis=0)
        )

        return [
            (parameters_min[index], parameters_max[index])
            for index in range(len(parameters_min))
        ]

    @property
    def log_evidence(self):
        return None



import numpy as np
import corner

from autofit.plot.samples_plotters import MCMCPlotter

class EmceePlotter(MCMCPlotter):

    def corner(self, **kwargs):

        corner.corner(
            data=np.asarray(self.samples.parameter_lists),
            weight_list=self.samples.weight_list,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="corner")
        self.close()

    def trajectories(self, **kwargs):

        self._plot_trajectories(
            samples=self.samples.results_internal.get_chain(),
            log_posterior_list=self.samples.results_internal.get_log_prob(),
            **kwargs
        )


    def likelihood_series(self, **kwargs):

        self._plot_likelihood_series(
            log_posterior_list = self.samples.results_internal.get_log_prob(),
            **kwargs
        )

    def time_series(self, **kwargs):

        self._plot_time_series(
            samples=self.samples.results_internal.get_chain(),
        )




from autoconf import conf
from autofit.text import formatter as frm, samples_text


def padding(item, target=6):
    string = str(item)
    difference = target - len(string)
    prefix = difference * " "
    return f"{prefix}{string}"


def result_info_from(samples) -> str:
    """
    Output the full model.results file, which include the most-likely model, most-probable model at 1 and 3
    sigma confidence and information on the maximum log likelihood.
    """

    results = []

    if hasattr(samples, "log_evidence"):
        if samples.log_evidence is not None:

            results += [
                frm.add_whitespace(
                    str0="Bayesian Evidence ",
                    str1="{:.8f}".format(samples.log_evidence),
                    whitespace=90
                )
            ]
            results += ["\n"]

    results += [
        frm.add_whitespace(
            str0="Maximum Log Likelihood ",
            str1="{:.8f}".format(max(samples.log_likelihood_list)),
            whitespace=90
        )
    ]
    results += ["\n"]
    results += [
        frm.add_whitespace(
            str0="Maximum Log Posterior ",
            str1="{:.8f}".format(max(samples.log_posterior_list)),
            whitespace=90
        )
    ]
    results += ["\n"]

    results += ["\n", samples.model.parameterization, "\n\n"]

    results += ["Maximum Log Likelihood Model:\n\n"]

    formatter = frm.TextFormatter()

    for i, prior_path in enumerate(samples.model.unique_prior_paths):
        formatter.add(
            prior_path, format_str().format(samples.max_log_likelihood_vector[i])
        )
    results += [formatter.text + "\n"]

    if hasattr(samples, "pdf_converged"):

        if samples.pdf_converged:

            results += samples_text.summary(samples=samples, sigma=3.0, indent=4, line_length=90)
            results += ["\n"]
            results += samples_text.summary(samples=samples, sigma=1.0, indent=4, line_length=90)

        else:

            results += [
                "\n WARNING: The samples have not converged enough to compute a PDF and model errors. \n "
                "The model below over estimates errors. \n\n"
            ]
            results += samples_text.summary(samples=samples, sigma=1.0, indent=4, line_length=90)

        results += ["\n\ninstances\n"]

    formatter = frm.TextFormatter()

    for t in samples.model.path_float_tuples:
        formatter.add(*t)

    results += ["\n" + formatter.text]

    return "".join(results)


def search_summary_from_samples(samples) -> [str]:
    line = [f"Total Samples = {samples.total_samples}\n"]
    if hasattr(samples, "total_accepted_samples"):
        line.append(f"Total Accepted Samples = {samples.total_accepted_samples}\n")
        line.append(f"Acceptance Ratio = {samples.acceptance_ratio}\n")
    if samples.time is not None:
        line.append(f"Time To Run = {samples.time}\n")
    return line


def search_summary_to_file(samples, log_likelihood_function_time, filename):
    summary = search_summary_from_samples(samples=samples)
    summary.append(f"Log Likelihood Function Evaluation Time (seconds) = {log_likelihood_function_time}")
    frm.output_list_of_strings_to_file(file=filename, list_of_strings=summary)


def format_str() -> str:
    """The format string for the model.results file, describing to how many decimal points every parameter
    estimate is output in the model.results file.
    """
    decimal_places = conf.instance["general"]["output"]["model_results_decimal_places"]
    return f"{{:.{decimal_places}f}}"



import logging
from collections import ChainMap
from numbers import Real
from typing import Dict, Tuple, Optional, Union, Iterable

import numpy as np

from autoconf import cached_property
from autofit import exc
from autofit.graphical.factor_graphs.abstract import AbstractNode
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.jacobians import AbstractJacobian
from autofit.graphical.utils import (
    StatusFlag,
    prod,
    OptResult,
    Status,
    LogWarnings,
)
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.collection import CollectionPriorModel
from autofit.mapper.variable import (
    Variable,
    Plate,
    VariableData,
    FactorValue,
)
from autofit.mapper.variable_operator import VariableFullOperator
from autofit.messages.abstract import AbstractMessage
from autofit.messages.fixed import FixedMessage
from autofit.messages.transform_wrapper import TransformedWrapperInstance

VariableFactorDist = Dict[str, Dict[Factor, AbstractMessage]]
Projection = Dict[str, AbstractMessage]

logger = logging.getLogger(__name__)

_log_projection_warnings = logger.debug


Delta = Union[float, "MeanField"]


def is_message(message):
    return isinstance(message, (AbstractMessage, TransformedWrapperInstance))


# Does this need to be a Factor?
class MeanField(CollectionPriorModel, Dict[Variable, AbstractMessage], Factor):
    """For a factor with multiple variables, this class represents the
    the mean field approximation to that factor,

    f(x₁, x₂, x₃) = q(x₁, x₂, x₃) = q₁(x₁) q₂(x₂) q₃(x₃)

    Internally these variables approximations are stored in a
    dictionary with the variables as keys and the message or
    variable distribution as values


    Methods
    -------
    keys()
        returns the variables of the mean_field

    logpdf({x₁: x1, x₂: x2, x₃: x3})
        returns the q(x₁, x₂, x₃), axis defines the axes over which
        to reduce the return, if the meanfield is duplicated over multiple
        plates.

    logpdf_gradient({x₁: x1, x₂: x2, x₃: x3})
        returns the q(x₁, x₂, x₃) and the gradients for each input.
        to save memory the gradients are always the shape of the input
        values (i.e. this does not calculate the Jacobian)
    """

    def __init__(
            self,
            dists: Dict[Variable, Union[AbstractMessage, float]],
            plates: Optional[Tuple[Plate, ...]] = None,
            log_norm: np.ndarray = 0.0,
    ):
        dict.__init__(self, dists)
        Factor.__init__(self, self._logpdf, *self, arg_names=[])
        CollectionPriorModel.__init__(self)

        if isinstance(dists, MeanField):
            self.log_norm = dists.log_norm
            self._plates = dists.plates
        else:
            self.log_norm = log_norm
            self._plates = self.sorted_plates if plates is None else plates

    def __radd__(self, other):
        return self + other

    def __add__(self, other):
        return type(self)(
            {
                variable: message + other
                for variable, message
                in self.items()
            },
            plates=self.plates,
            log_norm=self.log_norm,
        )

    def __rsub__(self, other):
        return (-self) + other

    def __sub__(self, other):
        return self + (-other)

    def __neg__(self):
        return type(self)(
            {
                variable: -message
                for variable, message
                in self.items()
            },
            plates=self.plates,
            log_norm=self.log_norm,
        )

    @cached_property
    def fixed_values(self):
        return VariableData(
            {k: dist.mean for k, dist in self.items() if isinstance(dist, FixedMessage)}
        )

    @classmethod
    def from_priors(cls, priors: Iterable[Prior]) -> "MeanField":
        """
        Create a MeanField from a list of priors.

        This works because priors are a kind of variable and
        messages can be derived from priors.

        Parameters
        ----------
        priors
            A list of priors

        Returns
        -------
        A mean field
        """
        return MeanField({prior: prior.message for prior in priors})

    pop = dict.pop
    values = dict.values
    items = dict.items
    __getitem__ = dict.__getitem__
    __len__ = dict.__len__

    @property
    def shapes(self):
        return {v: np.shape(m) for v, m in self.items()}

    @property
    def sizes(self):
        return {v: np.size(m) for v, m in self.items()}

    def __getitem__(self, index):
        if isinstance(index, Variable):
            return dict.__getitem__(self, index)
        else:
            return self.subset(self.variables, plates_index=index)

    def __setitem__(self, index, value):
        if isinstance(index, Variable):
            dict.__setitem__(self, index, value)
        elif isinstance(value, MeanField):
            self.update_mean_field(value, index)

    def merge(self, index, mean_field):
        new_dist = dict(self)
        if index:
            plate_sizes = VariableData.plate_sizes(self)
            for v, message in mean_field.items():
                i = v.make_indexes(index, plate_sizes)
                new_dist[v] = new_dist[v].merge(i, message)
        else:
            new_dist.update(mean_field)

        return MeanField(new_dist)

    def update_mean_field(self, mean_field, plates_index=None):
        if plates_index:
            plate_sizes = VariableData.plate_sizes(self)
            for v, new_message in mean_field.items():
                index = v.make_indexes(plates_index, plate_sizes)
                self[v][index] = new_message
        else:
            self.update(mean_field)

        return self

    def subset(self, variables=None, plates_index=None):
        cls = type(self) if isinstance(self, MeanField) else MeanField
        variables = variables or self.variables
        if plates_index:
            plate_sizes = VariableData.plate_sizes(self)
            variable_index = (
                (v, v.make_indexes(plates_index, plate_sizes)) for v in variables
            )
            mean_field = dict((v, self[v][index]) for v, index in variable_index)

            return cls(mean_field)

        return cls((v, self[v]) for v in variables if v in self.keys())

    def rescale(self, rescale: Dict[Variable, float]) -> "MeanField":
        rescaled = {}
        for v, message in self.items():
            scale = rescale.get(v, 1)
            if scale == 1:
                rescaled[v] = message
            elif scale == 0:
                rescaled[v] = 1.
            else:
                rescaled[v] = message ** scale

        return MeanField(rescaled)

    @property
    def mean(self):
        return VariableData({v: dist.mean for v, dist in self.items()})

    @property
    def variance(self):
        return VariableData({v: dist.variance for v, dist in self.items()})

    @property
    def std(self):
        return VariableData({v: dist.std for v, dist in self.items()})

    @property
    def scale(self):
        return VariableData({v: dist.scale for v, dist in self.items()})

    @property
    def lower_limit(self):
        return VariableData({
            v: np.full(m.shape, m.lower_limit) if m.shape else m.lower_limit
            for v, m in self.items()
        })

    @property
    def upper_limit(self):
        return VariableData({
            v: np.full(m.shape, m.upper_limit) if m.shape else m.upper_limit
            for v, m in self.items()
        })

    def precision(self, variables=None):
        variables = variables or self.all_variables
        variances = MeanField.variance.fget(self).subset(variables)
        return VariableFullOperator.from_diagonal(variances ** -1)

    @property
    def arguments(self) -> Dict[Variable, Prior]:
        """
        Arguments that can be used to update a PriorModel
        """
        return {v: dist for v, dist in self.items()}

    def _logpdf(self, *args: np.ndarray) -> np.ndarray:
        var_names = self.name_variable_dict
        return self.logpdf(dict(zip(self.args, args)))

    def logpdf(
            self,
            values: Dict[Variable, np.ndarray],
    ) -> np.ndarray:
        """Calculates the logpdf of the passed values for messages

        the result is broadcast to the appropriate shape given the variable
        plates
        """
        return sum(np.sum(m.logpdf(values[v])) for v, m in self.items())

    def logpdf_gradient(self, values: Dict[Variable, np.ndarray], **kwargs):
        logl = 0
        gradl = {}
        for v, m in self.items():
            lv, gradl[v] = m.logpdf_gradient(values[v])
            logl += np.sum(lv)

        return logl, gradl

    def __repr__(self):
        reprdict = (
                "{\n" + "\n".join(f"  {k}: {v}" for k, v in self.items()) + "\n  }"
        )
        classname = type(self).__name__
        return f"{classname}({reprdict}, log_norm={self.log_norm})"

    @property
    def is_valid(self):
        return all(d.is_valid for d in self.values())

    def prod(self, *approxs: "MeanField", default=None) -> "MeanField":
        dists = [
            (key, prod(
                (other_mean_field.get(key, 1.0) for other_mean_field in approxs),
                message)
             ) for key, message in self.items()
        ]
        if default is not None:
            dists = [
                (key, message if is_message(message) else default[key])
                for key, message in dists
            ]
        return MeanField({k: m for k, m in dists if is_message(m)})

    __mul__ = prod

    def __truediv__(self, other: "MeanField") -> "MeanField":
        return type(self)(
            {k: m / other.get(k, 1.0) for k, m in self.items()},
            self.log_norm - other.log_norm,
        )

    def __pow__(self, other: Union[float, "MeanField"]) -> "MeanField":
        if isinstance(other, Real):
            return type(self)(
                {k: m ** other for k, m in self.items()}, self.log_norm * other
            )
        return type(self)(
            {key: value ** other[key] for key, value in self.items()},
            self.log_norm * other.log_norm,
        )

    def log_normalisation(self, other: "MeanField") -> float:
        return sum(np.sum(dist.log_normalisation(other[v])) for v, dist in self.items())

    def update_invalid(self, other: "MeanField") -> "MeanField":
        mean_field = {}
        for k, m in self.items():
            m2 = other.get(k)
            mean_field[k] = m.update_invalid(m2) if m2 else m

        return type(self)(mean_field, self.log_norm)

    def check_valid(self):
        return VariableData((v, m.check_valid()) for v, m in self.items())

    def project_mode(self, res: OptResult):
        return self.from_mode_covariance(res.mode, res.hess_inv, res.log_norm)

    def from_opt_state(self, state):
        return self.from_mode_covariance(
            state.all_parameters, state.full_hessian.inv(), state.value
        )

    def from_mode_covariance(
            self,
            mode: Dict[Variable, np.ndarray],
            covar: Dict[Variable, np.ndarray],
            fun: Optional[float] = None,
    ):
        """
        Projects the mode and covariance
        """
        mode = ChainMap(mode, self.fixed_values)
        projection = MeanField(
            {
                v: self[v].from_mode(
                    mode[v],
                    covar.get(v),
                    id_=self[v].id,
                    lower_limit=self[v].lower_limit,
                    upper_limit=self[v].upper_limit,
                )
                for v in self.keys() & mode.keys()
            }
        )
        if fun is not None:
            projection.log_norm = fun - projection(mode).log_value

        return projection

    def sample(self, n_samples=None):
        return VariableData({v: dist.sample(n_samples) for v, dist in self.items()})

    def kl(self, mean_field: "MeanField") -> np.ndarray:
        return sum(np.sum(dist.kl(mean_field[k])) for k, dist in self.items())

    __hash__ = Factor.__hash__

    @classmethod
    def from_dist(
            cls, dist: Union[Dict[Variable, AbstractMessage], "MeanField"]
    ) -> "MeanField":
        return dist if isinstance(dist, cls) else MeanField(dist)

    def update_factor_mean_field(
            self,
            cavity_dist: "MeanField",
            last_dist: Optional["MeanField"] = None,
            delta: Delta = 1.0,
            status: Status = Status(),
    ) -> Tuple["MeanField", Status]:

        success, messages, _, flag = status
        updated = False
        try:
            with LogWarnings(logger=_log_projection_warnings, action='always') as caught_warnings:
                if isinstance(delta, MeanField) or delta < 1:
                    factor_dist = (self ** delta * last_dist ** (1 - delta)) / cavity_dist ** delta
                else:
                    factor_dist = self / cavity_dist

            for m in caught_warnings.messages:
                messages += (f"project_mean_field warning: {m}",)

            if not factor_dist.is_valid:
                success = False
                messages += (f"model projection for {self} is invalid",)
                factor_dist = factor_dist.update_invalid(last_dist)
                # May want to check another way
                # e.g. factor_dist.check_valid().sum() / factor_dist.check_valid().size
                valid = factor_dist.check_valid()
                if valid.any():
                    updated = True
                    n_valid = valid.sum()
                    n_total = valid.size
                    logger.debug(
                        "meanfield with variables: %r ,"
                        "partially updated %d parameters "
                        "out of %d total, %.0%%",
                        tuple(self.variables),
                        n_valid,
                        n_total,
                        n_valid / n_total,
                    )

                flag = StatusFlag.BAD_PROJECTION
            else:
                updated = True

        except exc.MessageException as e:
            logger.exception(e)
            factor_dist = last_dist

        return factor_dist, Status(
            success=success, messages=messages, updated=updated, flag=flag, result=status.result,
        )


class FactorApproximation(AbstractNode):
    """
    This class represents the 'tilted distribution' in EP,

    When approximating a model distribution of the form,

    m(x) = ∏ₐ fₐ(xₐ)

    we can define an approximating distribution as the product of
    factor distributions,

    q(x) = ∏ₐ qₐ(xₐ)

    the 'cavity distribution' q⁻ᵃ for a factor can be viewed as a
    prior distribution for the factor,

    q⁻ᵃ(xₐ) = ∏_{ᵦ ≠ a} qᵦ(xᵦ)

    so the model can be approximated by the 'tilted distribution'

    q⁺ᵃ(xₐ) = fₐ(xₐ) q⁻ᵃ(xₐ)

    Parameters
    ----------
    factor: Factor
        fₐ(xₐ)
    cavity_dist: MeanField
        q⁻ᵃ(xₐ)
    factor_dist: MeanField
        qₐ(xₐ)
    model_dist: MeanField
        q(xₐ)

    Methods
    -------
    __call__(values={xₐ: x₀})
        returns q⁺ᵃ(x₀)

    func_jacobian(values={xₐ: x₀}, variables=(xₐ,))
        returns q⁺ᵃ(x₀), {xₐ: dq⁺ᵃ(x₀)/dxₐ}

    project_mean_field(mean_field, delta=1., status=Status())
        for qᶠ = mean_field, finds qₐ such that qᶠₐ * q⁻ᵃ = qᶠ
        delta controls how much to change from the original factor qₐ
        so qʳₐ = (qᶠₐ)ᵟ * (qᶠₐ)¹⁻ᵟ

        returns qʳₐ, status
    """

    def __init__(
            self,
            factor: Factor,
            cavity_dist: MeanField,
            factor_dist: MeanField,
            model_dist: MeanField,
    ):
        # Have to seperate FactorApproximation into two classes
        # in order to be able to redefine __new__
        self.factor = factor
        self.cavity_dist = MeanField.from_dist(cavity_dist)
        self.factor_dist = MeanField.from_dist(factor_dist)
        self.model_dist = MeanField.from_dist(model_dist)

        self.fixed_values = self.factor_dist.fixed_values

        super().__init__(**factor._kwargs)

    @property
    def variables(self):
        return self.factor.variables

    @property
    def deterministic_variables(self):
        return self.factor.deterministic_variables

    @property
    def all_variables(self):
        return self.factor.all_variables

    @property
    def mean_field(self):
        return self.model_dist

    @property
    def name(self):
        return f"FactorApproximation({self.factor.name})"

    @property
    def deterministic_dist(self):
        """
        the `MeanField` approximation of the deterministic variables
        """
        return MeanField({v: self.cavity_dist[v] for v in self.deterministic_variables})

    def __call__(
            self,
            values: Dict[Variable, np.ndarray],
    ) -> FactorValue:
        variable_dict = {**self.fixed_values, **values}
        fval = self.factor(variable_dict)
        log_meanfield = self.cavity_dist({**variable_dict, **fval.deterministic_values})
        return np.sum(fval) + np.sum(log_meanfield)

    def func_jacobian(
            self, values: Dict[Variable, np.ndarray]
    ) -> Tuple[FactorValue, AbstractJacobian]:
        raise NotImplementedError()

    def func_gradient(
            self,
            values: Dict[Variable, np.ndarray],
    ) -> Tuple[FactorValue, VariableData]:
        variable_dict = {**self.fixed_values, **values}
        fval, fjac = self.factor.func_jacobian(variable_dict)

        variable_dict.update(fval.deterministic_values)
        log_cavity, grad_cavity = self.cavity_dist.logpdf_gradient(
            {**variable_dict, **fval.deterministic_values}
        )

        logl = np.sum(fval) + np.sum(log_cavity)
        grad = fjac.grad(grad_cavity)

        return logl, grad

    def project_mean_field(
            self,
            model_dist: MeanField,
            delta: Delta = 1.0,
            status: Status = Status(),
    ) -> Tuple["FactorApproximation", Status]:

        factor_dist, status = model_dist.update_factor_mean_field(
            self.cavity_dist,
            last_dist=self.factor_dist,
            delta=delta,
            status=status,
        )
        new_approx = FactorApproximation(
            self.factor,
            self.cavity_dist,
            factor_dist=factor_dist,
            model_dist=model_dist,
        )
        return new_approx, status

    project = project_mean_field

    def __repr__(self):
        # TODO make this nicer
        return f"{type(self).__name__}({self.factor}, ...)"



from autofit.plot import SamplesPlotter


class UltraNestPlotter(SamplesPlotter):

    def cornerplot(self, **kwargs):

        from ultranest import plot

        plot.cornerplot(
            results=self.samples.results_internal,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cornerplot")
        self.close()

    def runplot(self, **kwargs):

        from ultranest import plot

        try:
            plot.runplot(
                results=self.samples.results_internal,
                **kwargs
            )
        except KeyError:
            pass

        self.output.to_figure(structure=None, auto_filename="runplot")
        self.close()

    def traceplot(self, **kwargs):

        from ultranest import plot

        try:
            plot.traceplot(
                results=self.samples.results_internal,
                **kwargs
            )
        except KeyError:
            pass

        self.output.to_figure(structure=None, auto_filename="traceplot")
        self.close()


import logging
import os
from pathlib import Path
from typing import Dict, Optional, List

from autofit import exc
from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.expectation_propagation.history import EPHistory
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.graph import FactorGraph
from autofit.graphical.mean_field import Status
from autofit.graphical.utils import StatusFlag, LogWarnings
from autofit.mapper.identifier import Identifier
from autofit.non_linear.paths import DirectoryPaths
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.tools.util import IntervalCounter
from .factor_optimiser import AbstractFactorOptimiser, ExactFactorFit
from .visualise import Visualise

logger = logging.getLogger(__name__)


class EPOptimiser:
    def __init__(
            self,
            factor_graph: FactorGraph,
            default_optimiser: Optional[AbstractFactorOptimiser] = None,
            factor_optimisers: Optional[Dict[Factor, AbstractFactorOptimiser]] = None,
            ep_history: Optional[EPHistory] = None,
            factor_order: Optional[List[Factor]] = None,
            paths: AbstractPaths = None,
            factor_loggers: Optional[Dict[Factor, logging.Logger]] = None,
    ):
        """
        Optimise a factor graph.

        Cycles through factors optimising them individually using priors
        created through expectation propagation; in effect the prior for each
        variable for a given factor is the product of the posteriors of that
        variable for all other factors.

        Parameters
        ----------
        factor_graph
            A graph describing the relationships between multiple factors
        default_optimiser
            An optimiser that is used if no specific optimiser is provided for a factor
        factor_optimisers
            Specific optimisers used to optimise each factor
        ep_history
            Optionally specify an alternate history
        factor_order
            The factors in the graph but placed in the order in which they should
            be optimised
        paths
            Optionally define how data should be output
        """
        self.paths = paths or DirectoryPaths(identifier=str(Identifier(factor_graph)))

        factor_optimisers = factor_optimisers or {}
        self.factor_graph = factor_graph
        self.factors = factor_order or self.factor_graph.factors
        self.default_optimiser = default_optimiser

        if default_optimiser is None:
            self.factor_optimisers = factor_optimisers
            missing = set(self.factors) - self.factor_optimisers.keys()
            if missing:
                raise (
                    ValueError(
                        f"missing optimisers for {missing}, "
                        "pass a default_optimiser or add missing optimsers"
                    )
                )
        else:
            self.factor_optimisers = {
                factor: factor_optimisers.get(factor, default_optimiser)
                for factor in self.factors
            }

        for optimiser in self.factor_optimisers.values():
            optimiser.paths = self.paths

        self.ep_history = ep_history or EPHistory()
        self.factor_loggers = factor_loggers or {}

        with open(self.output_path / "graph.info", "w+") as f:
            f.write(self.factor_graph.info)

        self.visualiser = Visualise(self.ep_history, self.output_path)

    @classmethod
    def from_meanfield(
            cls,
            model_approx: EPMeanField,
            default_optimiser: Optional[AbstractFactorOptimiser] = None,
            factor_optimisers: Optional[Dict[Factor, AbstractFactorOptimiser]] = None,
            ep_history: Optional[EPHistory] = None,
            factor_order: Optional[List[Factor]] = None,
            paths: AbstractPaths = None,
            factor_loggers: Optional[Dict[Factor, logging.Logger]] = None,
            exact_fit_kws=None,
    ):
        factor_graph = model_approx.factor_graph
        factor_order = factor_order or factor_graph.factors

        factor_optimisers = factor_optimisers or {}
        factor_mean_field = model_approx.factor_mean_field
        for factor in factor_graph.factors:
            if factor not in factor_optimisers:
                factor_dist = factor_mean_field[factor]
                if factor.has_exact_projection(factor_dist):
                    factor_optimisers[factor] = ExactFactorFit(**(exact_fit_kws or {}))
                else:
                    factor_optimisers[factor] = default_optimiser

        return cls(
            factor_graph=factor_graph,
            default_optimiser=default_optimiser,
            factor_optimisers=factor_optimisers,
            ep_history=ep_history,
            factor_order=factor_order,
            paths=paths,
            factor_loggers=factor_loggers,
        )

    @property
    def output_path(self) -> Path:
        """
        The path at which data will be output. Uses the name of the optimiser.

        If the path does not exist it is created.
        """
        path = Path(self.paths.output_path)
        os.makedirs(path, exist_ok=True)
        return path

    def _log_factor(self, factor: Factor):
        """
        Log information for the factor and its history.
        """
        factor_logger = logging.getLogger(factor.name)
        try:
            factor_history = self.ep_history[factor]
            if factor_history.history:
                log_evidence = factor_history.latest_update.log_evidence
                divergence = factor_history.kl_divergence()

                factor_logger.info(f"Log Evidence = {log_evidence}")
                factor_logger.info(f"KL Divergence = {divergence}")
        except exc.HistoryException as e:
            factor_logger.exception(e)

    def factor_step(
            self,
            factor,
            model_approx,
            optimiser=None
    ):
        factor_logger = self.factor_loggers.get(factor.name, logging.getLogger(factor.name))
        factor_logger.debug("Optimising...")

        optimiser = optimiser or self.factor_optimisers[factor]
        try:
            with LogWarnings(logger=factor_logger.debug, action='always') as caught_warnings:
                model_approx, status = optimiser.optimise(
                    factor,
                    model_approx,
                )

            messages = status.messages + tuple(caught_warnings.messages)

            status = Status(status.success, messages, status.flag, result=status.result)

        except (ValueError, ArithmeticError, RuntimeError) as e:
            logger.exception(e)
            status = Status(
                False,
                (f"Factor: {factor} experienced error {e}",),
                StatusFlag.FAILURE,
            )

        factor_logger.debug(status)
        return model_approx, status

    def run(
            self,
            model_approx: EPMeanField,
            max_steps: int = 100,
            log_interval: int = 10,
            visualise_interval: int = 100,
            output_interval: int = 10,
    ) -> EPMeanField:
        """
        Run the optimisation on an approximation of the model.

        Parameters
        ----------
        model_approx
            A collection of messages describing priors on the model's variables.
        max_steps
            The maximum number of steps prior to termination. Termination may also
            occur when difference in log evidence or KL Divergence drop below a given
            threshold for two consecutive optimisations of a given factor.
        log_interval
            How steps should we wait before logging information?
        visualise_interval
            How steps should we wait before visualising information?
            This includes plots of KL Divergence and Evidence.
        output_interval
            How steps should we wait before outputting information?
            This includes the model.results file which describes the current mean values
            of each message.

        Returns
        -------
        An updated approximation of the model
        """
        should_log = IntervalCounter(log_interval)
        should_visualise = IntervalCounter(visualise_interval)
        should_output = IntervalCounter(output_interval)

        for _ in range(max_steps):
            _should_log = should_log()
            _should_visualise = should_visualise()
            _should_output = should_output()
            for factor, optimiser in self.factor_optimisers.items():
                model_approx, status = self.factor_step(
                    factor, model_approx, optimiser
                )
                if status and _should_log:
                    self._log_factor(factor)

                if self.ep_history(factor, model_approx, status):
                    logger.info("Terminating optimisation")
                    break  # callback controls convergence

            else:  # If no break do next iteration
                if _should_visualise:
                    self.visualiser()
                if _should_output:
                    self._output_results(model_approx)
                continue
            break  # stop iterations

        self.visualiser()
        self._output_results(model_approx)

        return model_approx

    def _output_results(self, model_approx: EPMeanField):
        """
        Save the graph.results text
        """
        with open(self.output_path / "graph.results", "w+") as f:
            f.write(self.factor_graph.make_results_text(model_approx))



from typing import Optional

from autofit.non_linear.samples import PDFSamples, Sample, NestSamples



def samples_with_log_likelihood_list(
        log_likelihood_list
):
    return [
        Sample(
            log_likelihood=log_likelihood,
            log_prior=0,
            weight=0
        )
        for log_likelihood
        in log_likelihood_list
    ]


class MockSamples(PDFSamples):
    def __init__(
            self,
            model=None,
            sample_list=None,
            max_log_likelihood_instance=None,
            log_likelihood_list=None,
            gaussian_tuples=None,
            unconverged_sample_size=10,
            **kwargs,
    ):

        self._log_likelihood_list = log_likelihood_list

        self.model = model

        sample_list = sample_list or self.default_sample_list

        super().__init__(
            model=model, sample_list=sample_list, unconverged_sample_size=unconverged_sample_size, **kwargs
        )

        self._max_log_likelihood_instance = max_log_likelihood_instance
        self._gaussian_tuples = gaussian_tuples

    @property
    def default_sample_list(self):

        if self._log_likelihood_list is not None:
            log_likelihood_list = self._log_likelihood_list
        else:
            log_likelihood_list = range(3)

        return [
            Sample(
                log_likelihood=log_likelihood,
                log_prior=0.0,
                weight=0.0
            )
            for log_likelihood
            in log_likelihood_list
        ]

    @property
    def log_likelihood_list(self):

        if self._log_likelihood_list is None:
            return super().log_likelihood_list

        return self._log_likelihood_list

    @property
    def max_log_likelihood_instance(self):

        if self._max_log_likelihood_instance is None:

            try:
                return super().max_log_likelihood_instance
            except (KeyError, AttributeError):
                pass

        return self._max_log_likelihood_instance

    def gaussian_priors_at_sigma(self, sigma=None):

        if self._gaussian_tuples is None:
            return super().gaussian_priors_at_sigma(sigma=sigma)

        return self._gaussian_tuples

    def write_table(self, filename):
        pass


class MockNestSamples(NestSamples):

    def __init__(
            self,
            model,
            sample_list=None,
            total_samples=10,
            log_evidence=0.0,
            number_live_points=5,
            time: Optional[float] = None,
    ):

        self.model = model

        if sample_list is None:

            sample_list = [
                Sample(
                    log_likelihood=log_likelihood,
                    log_prior=0.0,
                    weight=0.0
                )
                for log_likelihood
                in self.log_likelihood_list
            ]

        super().__init__(
            model=model,
            sample_list=sample_list,
            time=time
        )

        self._total_samples = total_samples
        self._log_evidence = log_evidence
        self._number_live_points = number_live_points

    @property
    def total_samples(self):
        return self._total_samples

    @property
    def log_evidence(self):
        return self._log_evidence

    @property
    def number_live_points(self):
        return self._number_live_points






from dynesty import plotting as dyplot

from autoconf import conf
from autofit.plot import SamplesPlotter



class DynestyPlotter(SamplesPlotter):

    def boundplot(self, **kwargs):

        if conf.instance["general"]["test"]["test_mode"]:
            return None

        dyplot.boundplot(
            results=self.samples.results_internal,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="boundplot")
        self.close()

    def cornerbound(self, **kwargs):

        if conf.instance["general"]["test"]["test_mode"]:
            return None

        dyplot.cornerbound(
            results=self.samples.results_internal,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cornerbound")
        self.close()

    def cornerplot(self, **kwargs):

        if conf.instance["general"]["test"]["test_mode"]:
            return None

        dyplot.cornerplot(
            results=self.samples.results_internal,
            labels=self.model.parameter_labels_with_superscripts_latex,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cornerplot")
        self.close()

    def cornerpoints(self, **kwargs):

        if conf.instance["general"]["test"]["test_mode"]:
            return None

        try:
            dyplot.cornerpoints(
                results=self.samples.results_internal,
                labels=self.model.parameter_labels_with_superscripts_latex,
                **kwargs
            )

            self.output.to_figure(structure=None, auto_filename="cornerpoints")
        except ValueError:
            pass

        self.close()

    def runplot(self, **kwargs):

        if conf.instance["general"]["test"]["test_mode"]:
            return None

        try:
            dyplot.runplot(
                results=self.samples.results_internal,
                **kwargs
            )
        except ValueError:
            pass

        self.output.to_figure(structure=None, auto_filename="runplot")
        self.close()

    def traceplot(self, **kwargs):

        if conf.instance["general"]["test"]["test_mode"]:
            return None

        dyplot.traceplot(
            results=self.samples.results_internal,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="traceplot")
        self.close()


import numpy as np

from autoconf import conf
from autofit import exc
from autofit.messages.normal import NormalMessage, UniformNormalMessage, LogNormalMessage
from autofit.messages.transform import log_10_transform
from autofit.messages.transform_wrapper import TransformedWrapperInstance
from .abstract import Prior
from .abstract import epsilon


class Limits:
    @staticmethod
    def for_class_and_attributes_name(cls, attribute_name):
        limit_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name, "gaussian_limits"]
        )
        return limit_dict["lower"], limit_dict["upper"]


class WrappedInstance(
    TransformedWrapperInstance
):
    __identifier_fields__ = ("lower_limit", "upper_limit")

    __database_args__ = (
        "lower_limit",
        "upper_limit",
        "log_norm",
        "id_",
    )

    def __init__(
            self,
            transformed_wrapper,
            *args,
            lower_limit,
            upper_limit,
            **kwargs
    ):
        super().__init__(
            transformed_wrapper,
            *args,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            **kwargs
        )
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

        if self.lower_limit >= self.upper_limit:
            raise exc.PriorException(
                "The upper limit of a prior must be greater than its lower limit"
            )

    def _new_for_base_message(
            self,
            message
    ):
        return type(self)(
            self.transformed_wrapper,
            *message.parameters,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
            id_=self.instance().id,
        )



import math
import warnings
from typing import Union

import numpy as np
from scipy.special import betaln
from scipy.special import psi, polygamma

from autoconf import cached_property
from ..messages.abstract import AbstractMessage


def grad_betaln(ab):
    psiab = psi(ab.sum(axis=1, keepdims=True))
    return psi(ab) - psiab


def jac_grad_betaln(ab):
    psi1ab = polygamma(1, ab.sum(axis=1, keepdims=True))
    fii = polygamma(1, ab) - psi1ab
    fij = -psi1ab[:, 0]
    return np.array([[fii[:, 0], fij], [fij, fii[:, 1]]]).T


def inv_beta_suffstats(lnX, ln1X):
    """Solve for a, b for,

    psi(a) + psi(a + b) = lnX
    psi(b) + psi(a + b) = ln1X
    """
    _lnX, _ln1X = np.ravel(lnX), np.ravel(ln1X)
    lnXs = np.c_[_lnX, _ln1X]

    # Find initial starting location
    Gs = np.exp(lnXs)
    dG = 1 - Gs.sum(axis=1, keepdims=True)
    ab = np.maximum(1, (1 + Gs / dG) / 2)

    # 5 Newton Raphson itertions is generally enough
    for i in range(5):
        f = grad_betaln(ab) - lnXs
        jac = jac_grad_betaln(ab)
        ab += np.linalg.solve(jac, - f)

    if np.any(ab < 0):
        warnings.warn(
            "invalid negative parameters found for inv_beta_suffstats, "
            "clampling value to 0.5",
            RuntimeWarning
        )
        b = np.clip(ab, 0.5, None)

    shape = np.shape(lnX)
    if shape:
        a = ab[:, 0].reshape(shape)
        b = ab[:, 1].reshape(shape)
    else:
        a, b = ab[0, :]

    return a, b


class BetaMessage(AbstractMessage):
    """
    Models a Beta distribution
    """
    log_base_measure = 0
    _support = ((0, 1),)
    _min = 0
    _max = 1
    _range = 1
    _parameter_support = ((0, np.inf), (0, np.inf))

    def __init__(
            self,
            alpha=0.5,
            beta=0.5,
            lower_limit=-math.inf,
            upper_limit=math.inf,
            log_norm=0,
            id_=None
    ):
        self.alpha = alpha
        self.beta = beta
        super().__init__(
            alpha,
            beta,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            log_norm=log_norm,
            id_=id_
        )

    def value_for(self, unit: float) -> float:
        raise NotImplemented()

    @cached_property
    def log_partition(self) -> np.ndarray:
        return betaln(*self.parameters)

    @cached_property
    def natural_parameters(self) -> np.ndarray:
        return self.calc_natural_parameters(
            self.alpha,
            self.beta
        )

    @staticmethod
    def calc_natural_parameters(
            alpha: Union[float, np.ndarray],
            beta: Union[float, np.ndarray]
    ) -> np.ndarray:
        return np.array([alpha - 1, beta - 1])

    @staticmethod
    def invert_natural_parameters(
            natural_parameters: np.ndarray
    ) -> np.ndarray:
        return natural_parameters + 1

    @classmethod
    def invert_sufficient_statistics(
            cls, sufficient_statistics: np.ndarray
    ) -> np.ndarray:
        a, b = inv_beta_suffstats(*sufficient_statistics)
        return cls.calc_natural_parameters(a, b)

    @classmethod
    def to_canonical_form(cls, x: np.ndarray) -> np.ndarray:
        return np.array([np.log(x), np.log1p(-x)])

    @cached_property
    def mean(self) -> Union[np.ndarray, float]:
        return self.alpha / (self.alpha + self.beta)

    @cached_property
    def variance(self) -> Union[np.ndarray, float]:
        return (
                self.alpha * self.beta
                / (self.alpha + self.beta) ** 2
                / (self.alpha + self.beta + 1)
        )

    def sample(self, n_samples=None):
        a, b = self.parameters
        shape = (n_samples,) + self.shape if n_samples else self.shape
        return np.random.beta(a, b, size=shape)

    def kl(self, dist):
        # TODO check this is correct
        # https://arxiv.org/pdf/0911.4863.pdf
        if self._support != dist._support:
            raise TypeError('Support does not match')

        aP, bP = dist.parameters
        aQ, bQ = self.parameters
        return (
                betaln(aQ, bQ) - betaln(aP, bP)
                - (aQ - aP) * psi(aP)
                - (bQ - bP) * psi(bP)
                + (aQ - aP + bQ - bP) * psi(aP + bP)
        )

    def logpdf_gradient(self, x):
        logl = self.logpdf(x)
        a, b = self.parameters
        gradl = (a - 1) / x + (b - 1) / (x - 1)
        return logl, gradl

    def logpdf_gradient_hessian(self, x):
        logl = self.logpdf(x)
        a, b = self.parameters
        ax, bx = (a - 1) / x, (b - 1) / (x - 1)
        gradl = ax + bx
        hessl = -ax / x - bx / (x - 1)
        return logl, gradl, hessl



from autofit.messages.normal import LogNormalMessage
from .abstract import Prior
from .wrapped_instance import WrappedInstance


class LogGaussianPrior(Prior):
    """A prior with a log gaussian distribution"""

    __identifier_fields__ = (
        "lower_limit",
        "upper_limit",
        "mean",
        "sigma"
    )

    def __init__(
            self,
            mean,
            sigma,
            lower_limit=0.0,
            upper_limit=float("inf"),
            id_=None,
    ):
        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)

        self.mean = mean
        self.sigma = sigma

        message = WrappedInstance(
            LogNormalMessage,
            mean=mean,
            sigma=sigma,
            id_=id_,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )
        super().__init__(
            message=message,
            lower_limit=0.0,
            upper_limit=float("inf"),
            id_=None,
        )

    def _new_for_base_message(
            self,
            message
    ):
        """
        Create a new instance of this wrapper but change the parameters used
        to instantiate the underlying message. This is useful for retaining
        the same transform stack after recreating the underlying message during
        projection.
        """
        return LogGaussianPrior(
            *message.parameters,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
            id_=self.instance().id
        )

    # @assert_within_limits
    def value_for(self, unit, ignore_prior_limits=False):
        """

        Parameters
        ----------
        unit: Float
            A unit hypercube value between 0 and 1

        Returns
        -------
        value: Float
            A value for the attribute biased to the gaussian distribution
        """
        return super().value_for(unit, ignore_prior_limits=ignore_prior_limits)



from abc import ABC, abstractmethod
from typing import List, Iterator

from .search_output import SearchOutput


class AttributePredicate:
    def __init__(self, *path):
        """
        Used to produce predicate objects for filtering in the aggregator.

        When an unrecognised attribute is called on an aggregator an instance
        of this object is created. This object implements comparison methods
        facilitating construction of predicates.

        Parameters
        ----------
        path
            A series of names of attributes that can be used to get a value.
            For example, (mask, pixel_size) would get the pixel size of a mask
            when evaluated for a given search.
        """
        self.path = path

    def value_for_search_output(
            self,
            search_output: SearchOutput
    ):
        """
        Recurse the search output by iterating the attributes in the path
        and getting a value for each attribute.
        """
        value = search_output
        for attribute in self.path:
            value = getattr(
                value,
                attribute
            )
        return value

    def __eq__(self, value):
        """
        Returns a predicate which asks whether the given value is equal to
        the attribute of a search.
        """
        return EqualityPredicate(
            self,
            value
        )

    def __le__(self, other):
        return OrPredicate(
            self == other,
            self < other
        )

    def __ge__(self, other):
        return OrPredicate(
            self == other,
            self > other
        )

    def __getattr__(self, item: str) -> "AttributePredicate":
        """
        Adds another item to the path
        """
        return AttributePredicate(
            *self.path, item
        )

    def __gt__(self, other):
        """
        Is the value of this attribute for a given search greater than some
        other value?
        """
        return GreaterThanPredicate(
            self, other
        )

    def __lt__(self, other):
        """
        Is the value of this attribute for a given search less than some
        other value?
        """
        return LessThanPredicate(
            self, other
        )

    def __ne__(self, other):
        """
        Returns a predicate which asks whether the given value is not equal to
        the attribute of a search.
        """
        return ~(self == other)

    def contains(self, value):
        """
        Returns a predicate which asks whether the given is contained within
        the attribute of a search.
        """
        return ContainsPredicate(
            self,
            value
        )


class AbstractPredicate(ABC):
    """
    Comparison between a value and some attribute of a search
    """

    def filter(
            self,
            search_outputs: List[SearchOutput]
    ) -> Iterator[SearchOutput]:
        """
        Only return searchs for which this predicate evaluates to True

        Parameters
        ----------
        search_outputs

        Returns
        -------

        """
        return filter(
            lambda search_output: self(search_output),
            search_outputs
        )

    def __invert__(self) -> "NotPredicate":
        """
        A predicate that evaluates to `True` when this predicate evaluates
        to False
        """
        return NotPredicate(
            self
        )

    def __or__(self, other: "AbstractPredicate") -> "OrPredicate":
        """
        Returns a predicate that is true if either predicate is true
        for a given search.
        """
        return OrPredicate(self, other)

    def __and__(self, other: "AbstractPredicate") -> "AndPredicate":
        """
        Returns a predicate that is true if both predicates are true
        for a given search.
        """
        return AndPredicate(self, other)

    @abstractmethod
    def __call__(self, search_output: SearchOutput) -> bool:
        """
        Does the attribute of the search match the requirement of this predicate?
        """


class CombinationPredicate(AbstractPredicate, ABC):
    def __init__(
            self,
            one: AbstractPredicate,
            two: AbstractPredicate
    ):
        """
        Abstract predicate combining two other predicates.

        Parameters
        ----------
        one
        two
            Child predicates
        """
        self.one = one
        self.two = two


class OrPredicate(CombinationPredicate):
    def __call__(self, search_output: SearchOutput):
        """
        The disjunction of two predicates.

        Parameters
        ----------
        search_output
            An object representing the output of a given search.

        Returns
        -------
        True if either predicate is `True` for the search
        """
        return self.one(search_output) or self.two(search_output)


class AndPredicate(CombinationPredicate):
    def __call__(self, search_output: SearchOutput):
        """
        The conjunction of two predicates.

        Parameters
        ----------
        search_output
            An object representing the output of a given search.

        Returns
        -------
        True if both predicates are `True` for the search
        """
        return self.one(search_output) and self.two(search_output)


class ComparisonPredicate(AbstractPredicate, ABC):
    def __init__(
            self,
            attribute_predicate: AttributePredicate,
            value
    ):
        """
        Compare an attribute of a search with a value.

        Parameters
        ----------
        attribute_predicate
            An attribute path of a search
        value
            A value to which the attribute is compared
        """
        self.attribute_predicate = attribute_predicate
        self._value = value

    def value(
            self,
            search_output
    ):
        if isinstance(self._value, AttributePredicate):
            return self._value.value_for_search_output(
                search_output
            )
        return self._value


class GreaterThanPredicate(ComparisonPredicate):
    def __call__(
            self,
            search_output: SearchOutput
    ) -> bool:
        """
        Parameters
        ----------
        search_output
            An object representing the output of a given search.

        Returns
        -------
        True iff the value of the attribute of the search is greater than
        the value associated with this predicate
        """

        return self.attribute_predicate.value_for_search_output(
            search_output
        ) > self.value(
            search_output
        )


class LessThanPredicate(ComparisonPredicate):
    def __call__(
            self,
            search_output: SearchOutput
    ) -> bool:
        """
        Parameters
        ----------
        search_output
            An object representing the output of a given search.

        Returns
        -------
        True iff the value of the attribute of the search is less than
        the value associated with this predicate
        """
        return self.attribute_predicate.value_for_search_output(
            search_output
        ) < self.value(
            search_output
        )


class ContainsPredicate(ComparisonPredicate):
    def __call__(
            self,
            search_output: SearchOutput
    ) -> bool:
        """
        Parameters
        ----------
        search_output
            An object representing the output of a given search.

        Returns
        -------
        True iff the value of the attribute of the search contains
        the value associated with this predicate
        """
        return self.value(
            search_output
        ) in self.attribute_predicate.value_for_search_output(
            search_output
        )


class EqualityPredicate(ComparisonPredicate):
    def __call__(self, search_output):
        """
        Parameters
        ----------
        search_output
            An object representing the output of a given search.

        Returns
        -------
        True iff the value of the attribute of the search is equal to
        the value associated with this predicate
        """
        try:
            value = self.value(
                search_output
            )
        except AttributeError:
            value = self.value
        return self.attribute_predicate.value_for_search_output(
            search_output
        ) == value


class NotPredicate(AbstractPredicate):
    def __init__(
            self,
            predicate: AbstractPredicate
    ):
        """
        Negates the output of a predicate.

        If the predicate would have returned `True` for a given search
        it now returns `False` and vice-versa.

        Parameters
        ----------
        predicate
            A predicate that is negated
        """
        self.predicate = predicate

    def __call__(self, search_output: SearchOutput) -> bool:
        """
        Evaluate the predicate for the search and return the negation
        of the result.

        Parameters
        ----------
        search_output
            The output of an AutoFit search

        Returns
        -------
        The negation of the underlying predicate
        """
        return not self.predicate(
            search_output
        )



from typing import Optional

from dynesty.dynesty import DynamicNestedSampler

from .abstract import AbstractDynesty, prior_transform


class DynestyDynamic(AbstractDynesty):
    __identifier_fields__ = (
        "bound",
        "sample",
        "enlarge",
        "bootstrap",
        "vol_dec",
        "walks",
        "facc",
        "slices",
        "fmove",
        "max_move"
    )

    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer=None,
            iterations_per_update: int = None,
            number_of_cores: int = None,
            **kwargs
    ):
        """
        A Dynesty non-linear search, using a dynamically changing number of live points.

        For a full description of Dynesty, checkout its GitHub and readthedocs webpages:

        https://github.com/joshspeagle/dynesty
        https://dynesty.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        iterations_per_update
            The number of iterations performed between every Dynesty back-up (via dumping the Dynesty instance as a
            pickle).
        number_of_cores
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            **kwargs
        )

        self.logger.debug("Creating DynestyDynamic Search")

    def sampler_from(
            self,
            model,
            fitness_function,
            pool=None
    ):
        """
        Get the dynamic Dynesty sampler which performs the non-linear search, passing it all associated input Dynesty
        variables.
        """

        live_points = self.live_points_from_model_and_fitness_function(
            model=model, fitness_function=fitness_function
        )

        return DynamicNestedSampler(
            loglikelihood=fitness_function,
            prior_transform=prior_transform,
            ndim=model.prior_count,
            logl_args=[model, fitness_function],
            ptform_args=[model],
            live_points=live_points,
            queue_size=self.number_of_cores,
            pool=pool,
            **self.config_dict_search
        )

    @property
    def total_live_points(self):
        return self.config_dict_run["nlive_init"]



from copy import copy

from autofit.mapper.prior_model import abstract
from autofit.non_linear.mock.mock_samples import MockSamples
from autofit.non_linear import abstract_search
from autofit.non_linear import paths
from autofit.non_linear.result import Result
from autofit.non_linear.grid.grid_search import make_lists


class GridSearch:

    def __init__(self, step_size=0.5, name=None):
        self.step_size = step_size
        self.paths = paths.DirectoryPaths(
            name=name
        )

    @property
    def name(self):
        return self.paths.name

    def copy_with_paths(self, paths):
        search = copy(self)
        search.paths = paths
        return search

    def fit(
            self,
            model: abstract.AbstractPriorModel,
            analysis: abstract_search.Analysis
    ):
        best_likelihood = float("-inf")
        best_instance = None

        likelihoods = list()

        for list_ in make_lists(
                no_dimensions=model.prior_count,
                step_size=self.step_size
        ):
            instance = model.instance_from_unit_vector(
                list_
            )
            likelihood = analysis.log_likelihood_function(
                instance
            )
            likelihoods.append(likelihood)
            if likelihood > best_likelihood:
                best_likelihood = likelihood
                best_instance = instance

        return Result(
            samples=MockSamples(
                max_log_likelihood_instance=best_instance,
                log_likelihood_list=likelihoods,
                gaussian_tuples=None
            ),
            model=model
        )
